#!python

# some forward-compatibilty functions for those unfortunate
#  souls that can't yet run Python 2.4

# attrgetter
try:
    from operator import attrgetter
except ImportError:
    def attrgetter(attr_name):
        return lambda target: getattr(target, attr_name)

# set
try:
    set = set
    frozenset = frozenset
except NameError:
    from sets import Set as set, ImmutableSet as frozenset

# sorted
try:
    sorted = sorted
except NameError:
    # global name 'sorted' doesn't exist in Python2.3
    # this provides a poor-man's emulation of the sorted built-in method
    def sorted(l, cmp=cmp, key=None, reverse=False):
        identity = lambda x: x
        if key is None:
            key = identity
        if reverse:
            cmp_ = lambda a,b: -cmp(a,b)
        else:
            cmp_ = cmp
        sorted_list = list(l)
        sorted_list.sort(
                lambda self, other: cmp_(
                        key(self),
                        key(other)))
        return sorted_list

# reversed
try:
    reversed = reversed
except NameError:
    def reversed(sequence):
        for item in sequence[::-1]:
            yield item

# rsplit
try:
    ''.rsplit
    def rsplit(s, delim, maxsplit):
        return s.rsplit(delim, maxsplit)

except AttributeError:
    def rsplit(s, delim, maxsplit):
        """Return a list of the words of the string s, scanning s
        from the end. To all intents and purposes, the resulting
        list of words is the same as returned by split(), except
        when the optional third argument maxsplit is explicitly
        specified and nonzero. When maxsplit is nonzero, at most
        maxsplit number of splits - the rightmost ones - occur,
        and the remainder of the string is returned as the first
        element of the list (thus, the list will have at most
        maxsplit+1 elements). New in version 2.4.
        >>> rsplit('foo.bar.baz', '.', 0)
        ['foo.bar.baz']
        >>> rsplit('foo.bar.baz', '.', 1)
        ['foo.bar', 'baz']
        >>> rsplit('foo.bar.baz', '.', 2)
        ['foo', 'bar', 'baz']
        >>> rsplit('foo.bar.baz', '.', 99)
        ['foo', 'bar', 'baz']
        """
        assert maxsplit >= 0

        if maxsplit == 0: return [s]

        # the following lines perform the function, but inefficiently.
        #  This may be adequate for compatibility purposes
        items = s.split(delim)
        if maxsplit < len(items):
            items[:-maxsplit] = [delim.join(items[:-maxsplit])]
        return items

try:
    from subprocess import list2cmdline
except ImportError:
    # here's list2cmdline copied directly from python 2.5.4
    def list2cmdline(seq):
        """
        Translate a sequence of arguments into a command line
        string, using the same rules as the MS C runtime:

        1) Arguments are delimited by white space, which is either a
           space or a tab.

        2) A string surrounded by double quotation marks is
           interpreted as a single argument, regardless of white space
           contained within.  A quoted string can be embedded in an
           argument.

        3) A double quotation mark preceded by a backslash is
           interpreted as a literal double quotation mark.

        4) Backslashes are interpreted literally, unless they
           immediately precede a double quotation mark.

        5) If backslashes immediately precede a double quotation mark,
           every pair of backslashes is interpreted as a literal
           backslash.  If the number of backslashes is odd, the last
           backslash escapes the next double quotation mark as
           described in rule 3.
        """

        # See
        # http://msdn.microsoft.com/library/en-us/vccelng/htm/progs_12.asp
        result = []
        needquote = False
        for arg in seq:
            bs_buf = []

            # Add a space to separate this argument from the others
            if result:
                result.append(' ')

            needquote = (" " in arg) or ("\t" in arg) or arg == ""
            if needquote:
                result.append('"')

            for c in arg:
                if c == '\\':
                    # Don't know if we need to double yet.
                    bs_buf.append(c)
                elif c == '"':
                    # Double backspaces.
                    result.append('\\' * len(bs_buf)*2)
                    bs_buf = []
                    result.append('\\"')
                else:
                    # Normal char
                    if bs_buf:
                        result.extend(bs_buf)
                        bs_buf = []
                    result.append(c)

            # Add remaining backspaces, if any.
            if bs_buf:
                result.extend(bs_buf)

            if needquote:
                result.extend(bs_buf)
                result.append('"')

        return ''.join(result)

try:
    from hashlib import md5
    from hashlib import sha1 as sha
except ImportError:
    from md5 import md5
    from sha import sha

try:
    next = next
except NameError:
    def next(iterable):
        return iterable.next()

def dictupdate(target, *args, **kwargs):
    target.update(dict(*args, **kwargs))

import sys
if sys.version_info >= (2,4):
    from xlutils.copy import copy as xlutils_copy
else:
    # xlutils expects 'set' in builtins, so put it there
    import __builtin__
    __builtin__.set = set
    from xlutils.copy import copy as xlutils_copy
    # gotta leave set in there or the module doesn't work.
    #del __builtin__.set
    del __builtin__

try:
    from operator import itemgetter
except ImportError:
    def itemgetter(*args):
        def callback(obj):
            return [obj[item] for item in args]
        return callback

if sys.version_info >= (2,4):
    from textwrap import dedent
else:
    # Python 2.3 dedent has a bug that removes tabs, so encode them
    from textwrap import dedent as dedent_orig
    def dedent(string):
        string = string.replace('\t', '__x0009__')
        return dedent_orig(string).replace('__x0009__', '\t')

try:
    from itertools import product
except ImportError:
    # from Python 2.6 docs, a memory-inefficient but equivalent algorithm
    def product(*args, **kwds):
        # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
        # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
        pools = map(tuple, args) * kwds.get('repeat', 1)
        result = [[]]
        for pool in pools:
            result = [x+[y] for x in result for y in pool]
        for prod in result:
            yield tuple(prod)

try:
    from itertools import chain
    chain_from_iterable = chain.from_iterable
except AttributeError:
    #from Python 2.6 docs
    def chain_from_iterable(iterables):
        # chain.from_iterable(['ABC', 'DEF']) --> A B C D E F
        for it in iterables:
            for element in it:
                yield element

try:
    from collections import deque
except ImportError:
    # A partial implementation of a deque
    class deque(list):
        def popleft(self):
            return self.pop(0)

try:
    from itertools import tee
except ImportError:
    # from Python 2.6 docs
    def tee(iterable, n=2):
        it = iter(iterable)
        deques = [deque() for i in range(n)]
        def gen(mydeque):
            while True:
                if not mydeque:             # when the local deque is empty
                    newval = next(it)       # fetch a new value and
                    for d in deques:        # load it to all the deques
                        d.append(newval)
                yield mydeque.popleft()
        return tuple([gen(d) for d in deques])

try:
    any = any
except NameError:
    # from Python 2.6 docs
    def any(iterable):
        for element in iterable:
            if element:
                return True
        return False

import logging
if 'level' in logging.basicConfig.func_code.co_varnames:
	logging_basicConfig = logging.basicConfig
else:
	def logging_basicConfig(**kwargs):
		logging.basicConfig()
		if 'level' in kwargs:
			logging.root.level = kwargs.pop('level')
		if kwargs:
			pass # ignore other Py2.4 parameters for now
del logging

if sys.version_info >= (2,6):
	def ljust(s, width, fillchar=' '):
		# the Python 3.0 way of left-justifying a string
		return '{0:{2}<{1}}'.format(s, width, fillchar)
else:
	# from implementation found in CheckLibs
	def ljust(src, width, fillchar=' '):
		if len(src) >= length:
			return src
		return src+(fillchar*(length-len(src)))

del sys



from py27compat import properties


def test_AutoBindingMethod():
	class Example(object):
		@properties.simplemethod
		def method(self):
			return 'called with {!r}'.format(self)

	ex = Example()
	assert ex.method().startswith('called with {!r}'.format(ex))

	assert Example.method('some value') == "called with 'some value'"




def format(str, *args, **kwargs):
	"""
	A very cheap hack to provide a basic implementation of
	string.format under Python 2.5
	"""
	if args:
		raise NotImplementedError("args not supported yet")
	try:
		return str.format(*args, **kwargs)
	except AttributeError:
		str = str.replace('{', '%(').replace('}', ')')
		return str % kwargs



from __future__ import absolute_import

import sys
import traceback

# On Python 2.7, format_exc always returns a UTF-8-encoded bytestring; make it
#  return a unicode string as it does on Python 3.
# I determined this finding experimentally. If you find that's not the case,
#  provide a counter example. Thanks.
def _format_exc_2x(*args, **kwargs):
	return traceback.format_exc(*args, **kwargs).decode('utf-8')

format_exc = (
	_format_exc_2x
	if sys.version_info < (3,) else
	traceback.format_exc
)



import types
import sys


class simplemethod(object):
	"""
	On Python 2, allow a method to bind to an instance
	(or not), just as it does on Python 3. Avoids TypeError
	when a bound method is called with a non-instance of
	the owner.
	"""
	def __new__(cls, func):
		if sys.version_info >= (3,):
			return func
		return super(simplemethod, cls).__new__(cls, func)

	def __init__(self, func):
		self.func = func

	def __get__(self, instance, owner):
		if instance is None:
			return self.func
		return types.MethodType(self.func, instance, owner)



"""
Module that provides backward-compatible implementations of Python 2.5
built-in capability. Named py24compat because when you no longer need Python
2.4 compatibility, you can remove py24compat.
"""
import sys

def _partition_py24(s, sep):
	""" Returns a three element tuple, (head, sep, tail) where:

		head + sep + tail == s
		sep == '' or sep is t
		bool(sep) == (t in s)		# sep indicates if the string was found

	>>> s = 'http://www.python.org'
	>>> partition(s, '://')
	('http', '://', 'www.python.org')
	>>> partition(s, '?')
	('http://www.python.org', '', '')
	>>> partition(s, 'http://')
	('', 'http://', 'www.python.org')
	>>> partition(s, 'org')
	('http://www.python.', 'org', '')
	"""
	if not isinstance(t, basestring) or not t:
		raise ValueError('partititon argument must be a non-empty string')
	parts = s.split(sep, 1)
	if len(parts) == 1:
		result = (s, '', '')
	else:
		result = (parts[0], sep, parts[1])
	assert len(result) == 3
	assert ''.join(result) == s
	assert result[1] == '' or result[1] is sep
	return result

if sys.version_info >= (2,5):
	partition = lambda s, sep: s.partition(sep)
else:
	partition = _partition_py24



from __future__ import absolute_import

import functools
from collections import namedtuple

try:
	import builtins
except ImportError:
	import __builtin__ as builtins

try:
	from thread import allocate_lock as Lock
except ImportError:
	try:
		from _thread import allocate_lock as Lock
	except:
		from _dummy_thread import allocate_lock as Lock

from py26compat.collections import OrderedDict
from ._functools_wraps import wraps

def _move_to_end(odict, key):
	'''
	Move an existing element to the end.

	Raises KeyError if the element does not exist.
	'''
	odict[key] = odict.pop(key)

# lru_cache implementation from Python 3.3dev [dfffb293f4b3]
_CacheInfo = namedtuple("CacheInfo", "hits misses maxsize currsize")

def lru_cache(maxsize=100, typed=False):
	"""Least-recently-used cache decorator.

	If *maxsize* is set to None, the LRU features are disabled and the cache
	can grow without bound.

	If *typed* is True, arguments of different types will be cached separately.
	For example, f(3.0) and f(3) will be treated as distinct calls with
	distinct results.

	Arguments to the cached function must be hashable.

	View the cache statistics named tuple (hits, misses, maxsize, currsize) with
	f.cache_info().	 Clear the cache and statistics with f.cache_clear().
	Access the underlying function with f.__wrapped__.

	See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used

	"""
	# Users should only access the lru_cache through its public API:
	#		cache_info, cache_clear, and f.__wrapped__
	# The internals of the lru_cache are encapsulated for thread safety and
	# to allow the implementation to change (including a possible C version).

	def decorating_function(user_function, **kwargs):
		tuple=kwargs.get('tuple', builtins.tuple)
		sorted=kwargs.get('sorted', builtins.sorted)
		map=kwargs.get('map', builtins.map)
		len=kwargs.get('len', builtins.len)
		type=kwargs.get('type', builtins.type)
		KeyError=kwargs.get('KeyError', builtins.KeyError)

		hits = [0]
		misses = [0]
		kwd_mark = (object(),)			# separates positional and keyword args
		lock = Lock()					# needed because OrderedDict isn't threadsafe

		if maxsize is None:
			cache = dict()				# simple cache without ordering or size limit

			@wraps(user_function)
			def wrapper(*args, **kwds):
				#nonlocal hits, misses
				key = args
				if kwds:
					sorted_items = tuple(sorted(kwds.items()))
					key += kwd_mark + sorted_items
				if typed:
					key += tuple(map(type, args))
					if kwds:
						key += tuple(type(v) for k, v in sorted_items)
				try:
					result = cache[key]
					hits[0] += 1
					return result
				except KeyError:
					pass
				result = user_function(*args, **kwds)
				cache[key] = result
				misses[0] += 1
				return result
		else:
			cache = OrderedDict()			# ordered least recent to most recent
			cache_popitem = cache.popitem
			# use the move_to_end method if available, otherwise fallback to
			#  the function.
			cache_renew = getattr(cache, 'move_to_end',
				functools.partial(_move_to_end, cache))

			@wraps(user_function)
			def wrapper(*args, **kwds):
				#nonlocal hits, misses
				key = args
				if kwds:
					sorted_items = tuple(sorted(kwds.items()))
					key += kwd_mark + sorted_items
				if typed:
					key += tuple(map(type, args))
					if kwds:
						key += tuple(type(v) for k, v in sorted_items)
				with lock:
					try:
						result = cache[key]
						cache_renew(key)	# record recent use of this key
						hits[0] += 1
						return result
					except KeyError:
						pass
				result = user_function(*args, **kwds)
				with lock:
					cache[key] = result		# record recent use of this key
					misses[0] += 1
					if len(cache) > maxsize:
						cache_popitem(0)	# purge least recently used cache entry
				return result

		def cache_info():
			"""Report cache statistics"""
			with lock:
				return _CacheInfo(hits[0], misses[0], maxsize, len(cache))

		def cache_clear():
			"""Clear the cache and cache statistics"""
			#nonlocal hits, misses
			with lock:
				cache.clear()
				hits[0] = misses[0] = 0

		wrapper.cache_info = cache_info
		wrapper.cache_clear = cache_clear
		return wrapper

	return decorating_function



# supplies OrderedDict via the ordereddict package for Python 2.5+
from __future__ import absolute_import

try:
	from collections import OrderedDict
except ImportError:
	from ordereddict import OrderedDict



"""
Python added .terminate() and .kill() to subprocess.Popen objects in
Python 2.6. If you need Python 2.5 compatibility, use term_proc and
kill_proc instead.

from py25compat.subprocess import term_proc

my_proc = Popen(...)
# do stuff...
term_proc(my_proc)
# or if you really need to kill it
kill_proc(my_proc)

"""

import sys
import os

def os_term(pid):
	import signal
	os.kill(pid, signal.SIGTERM)

def term_pid_win32(pid):
	raise NotImplementedError()

def os_kill(pid):
	import signal
	os.kill(pid, signal.SIGKILL)

def kill_pid_win32(pid):
	raise NotImplementedError()

if sys.version_info < (2,6):
	def term_proc(popen):
		"""
		Terminate a popen process
		"""
		func = globals().get('term_pid_' + sys.platform, os_term)
		func(popen.pid)
	def kill_proc(popen):
		"""
		Kill a popen process
		"""
		func = globals().get('kill_pid_' + sys.platform, os_kill)
		func(popen.pid)
else:
	def term_proc(popen):
		return popen.terminate()
	def kill_proc(popen):
		return popen.kill()



# this module provides forward-compatibility for Python 2.6 and 2.7
#  features on Python 2.5

try:
	import builtins
except ImportError:
	import __builtin__ as builtins

# next statement
try:
	next = next
except NameError:
	class __NotSupplied(object): pass
	def next(iterable, default=__NotSupplied):
		try:
			return iterable.next()
		except StopIteration:
			if default is __NotSupplied:
				raise
			return default

# namedtuple
try:
	from collections import namedtuple
except ImportError:
	from namedtuple_backport import namedtuple

# enumerate with start param
try:
	enumerate('abc', 1)
	enumerate = enumerate
except TypeError:
	def enumerate(seq, start=0):
		for index, item in builtins.enumerate(seq):
			yield index + start, item



from __future__ import absolute_import

import subprocess

try:
	check_output = subprocess.check_output
except AttributeError:
	# copied from Python 2.7.2
	def check_output(*popenargs, **kwargs):
		r"""Run command with arguments and return its output as a byte string.

		If the exit code was non-zero it raises a CalledProcessError.  The
		CalledProcessError object will have the return code in the returncode
		attribute and output in the output attribute.

		The arguments are the same as for the Popen constructor.  Example:

		>>> check_output(["ls", "-l", "/dev/null"])
		'crw-rw-rw- 1 root root 1, 3 Oct 18  2007 /dev/null\n'

		The stdout argument is not allowed as it is used internally.
		To capture standard error in the result, use stderr=STDOUT.

		>>> check_output(["/bin/sh", "-c",
		...               "ls -l non_existent_file ; exit 0"],
		...              stderr=STDOUT)
		'ls: non_existent_file: No such file or directory\n'
		"""
		if 'stdout' in kwargs:
			raise ValueError('stdout argument not allowed, it will be overridden.')
		process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
		output, unused_err = process.communicate()
		retcode = process.poll()
		if retcode:
			cmd = kwargs.get("args")
			if cmd is None:
				cmd = popenargs[0]
			err = subprocess.CalledProcessError(retcode, cmd)
			err.output = output
			raise err
		return output



#!/usr/bin/env python

# Project skeleton maintained at https://github.com/jaraco/skeleton

import io

import setuptools

with io.open('README.rst', encoding='utf-8') as readme:
	long_description = readme.read()

name = 'jaraco.compat'
description = 'Modules providing forward compatibility across some Python 2.x versions'

pre_27 = (
	':python_version == "2.6" or '
	'python_version == "2.5" or '
	'python_version == "2.4" or '
	'python_version == "2.3"'
)

params = dict(
	name=name,
	use_scm_version=True,
	author="Jason R. Coombs",
	author_email="jaraco@jaraco.com",
	description=description or name,
	long_description=long_description,
	url="https://github.com/jaraco/" + name,
	packages=setuptools.find_packages(),
	include_package_data=True,
	install_requires=[
		'six',
	],
	extras_require={
		pre_27: [
			'ordereddict >= 1.1',
		],
	},
	setup_requires=[
		'setuptools_scm>=1.15.0',
	],
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 2.3",
		"Programming Language :: Python :: 2.4",
		"Programming Language :: Python :: 2.5",
		"Programming Language :: Python :: 2.6",
		"Programming Language :: Python :: 2.7",
		"Programming Language :: Python :: 3",
	],
	entry_points={
	},
	use_2to3=True,
)
if __name__ == '__main__':
	setuptools.setup(**params)



from __future__ import absolute_import
import itertools

__all__ = ['product', 'izip_longest']

try:
	from itertools import product
except ImportError:
	# from Python 2.7 docs
	def product(*args, **kwds):
		# product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
		# product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
		pools = map(tuple, args) * kwds.get('repeat', 1)
		result = [[]]
		for pool in pools:
			result = [x+[y] for x in result for y in pool]
		for prod in result:
			yield tuple(prod)

try:
	from itertools import izip_longest
except ImportError:
	# from Python 2.7.1 docs
	def izip_longest(*args, **kwds):
		# izip_longest('ABCD', 'xy', fillvalue='-') --> Ax By C- D-
		fillvalue = kwds.get('fillvalue')
		def sentinel(counter = ([fillvalue]*(len(args)-1)).pop):
			yield counter()         # yields the fillvalue, or raises IndexError
		fillers = itertools.repeat(fillvalue)
		iters = [itertools.chain(it, sentinel(), fillers) for it in args]
		try:
			for tup in itertools.izip(*iters):
				yield tup
		except IndexError:
			pass



from __future__ import absolute_import

import functools

# Add the automatic addition of the __wrapped__ attribute when calling
#  update_wrapper or wraps.
def update_wrapper(wrapper, wrapped, *args, **kwargs):
	res = functools.update_wrapper(wrapper, wrapped, *args, **kwargs)
	res.__wrapped__ = wrapped
	return res

def wraps(wrapped, *args, **kwargs):
	return functools.partial(update_wrapper, wrapped=wrapped, *args, **kwargs)



from __future__ import division

try:
	from functools import cmp_to_key
except ImportError:
	# from Python 2.7 docs
	def cmp_to_key(mycmp):
		'Convert a cmp= function into a key= function'
		class K(object):
			def __init__(self, obj, *args):
				self.obj = obj
			def __lt__(self, other):
				return mycmp(self.obj, other.obj) < 0
			def __gt__(self, other):
				return mycmp(self.obj, other.obj) > 0
			def __eq__(self, other):
				return mycmp(self.obj, other.obj) == 0
			def __le__(self, other):
				return mycmp(self.obj, other.obj) <= 0
			def __ge__(self, other):
				return mycmp(self.obj, other.obj) >= 0
			def __ne__(self, other):
				return mycmp(self.obj, other.obj) != 0
		return K

def total_seconds(td):
	"""
	Python 2.7 adds a total_seconds method to timedelta objects.
	See http://docs.python.org/library/datetime.html#datetime.timedelta.total_seconds
	"""
	try:
		result = td.total_seconds()
	except AttributeError:
		result = (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
	return result
